# @admin.register(Rebate)
# class RebateModelAdmin(admin.ModelAdmin):
#     def has_add_permission(self, request, obj=None):
#         return False

#     def has_delete_permission(self, request, obj=None):
#         return False
